#!palmtcl

# palmtcl script to dump a datebook file into the format needed
# for bigcal (http://tony.aiu.to/sa/bigcal.html)
#

# Get command line stuff
set files "$argv"
#puts "args = $files"

if { [ info exists VERBOSE ] == 0 } {
        set VERBOSE 0
        catch { set VERBOSE $env(VERBOSE) } 
}
if { [ info exists QUIET ] == 0 } {
        set QUIET 0
        catch { set QUIET $env(QUIET) } 
}
if { $QUIET > 0 } {
        set VERBOSE 0
}

# If we are in tclsh, this will fail and we will load the package.
# When linked as palmtcl, it is true so we don't load as a package.

if { [info commands palm ] == "" } {
	set x [ catch  {
		package require palm
	} msg ] 
	if { "$x" == 0 } {
		puts "#loaded package: palm $msg"
	} else {
		puts "#Could not load palm package"
		exit 1
	}
}

if { 1 } {
puts "#Getting new datebook"
set res [exec /usr/bin/smbclient {\\windowshost\c} PASSWORD -U USER -D palm/USER/datebook -c {get datebook.dat}]

if { ! [regexp {.*getting file datebook.dat of size.*} $res ] } {
	puts "#Could not read datebook.dat from PC.  smb session: $res"
	exit 1
}

}



# some constants
set forever 0x749d3450
set secondsperday [expr 60 * 60 * 24]






proc time2str { timeval } {
	if { $timeval == 0x749d3450 } {
		return "forever";
	}
	return "[clock format $timeval -format {%Y-%02m-%02d %H:%M:%S} ]=$timeval"
}


proc printRepeat { e } {

	set eventFlag [$e eventFlag]
	if { $eventFlag == 0 } return

	puts -nonewline "#repeat {"
	if { $eventFlag == 0xffff } {
		puts -nonewline "class '[$e classname]'"
	} else {
		puts -nonewline "[format {0x%x} $eventFlag]"
	}
	set brand [$e brand]

	puts -nonewline " brand 0x[format %x $brand] interval [$e interval]"
	if { [$e endDate] != 0 } {
		puts -nonewline " enddate [time2str [$e endDate]]"
	}
	puts -nonewline " dow [$e dayOfWeek]"

	# brand data
	if { $brand <= 3 } {
		puts -nonewline " day [$e day]"
	} 
	if { $brand == 2 } {
		puts -nonewline " daymask [$e daymask]"
	}
	if { $brand == 3 } {
		puts -nonewline " weeki [$e weeki]"
	}
	if { $brand == 4 || $brand == 5 } {
		puts -nonewline " day# [$e day]"
	}
	if { $brand == 5 } {
		puts -nonewline " monthi [$e monthi]"
	}
	puts "}";
	if { [e nDateExceptions] > 0 } {
		puts -nonewline "#exceptions {"
		for { set di 0} {$di < [e nDateExceptions]} {incr di} {
			puts -nonewline " [e exceptions $di]"
		}
		puts "}"
	}
}







proc printMe {e day} {
	puts -nonewline "[clock format $day -format {%m %d} ]	"

	if { [regexp -nocase {birthday} [$e name]] } {
		puts -nonewline {+b }
	}
	if { [regexp -nocase {anniversay} [$e name]] } {
		puts -nonewline {+b }
	}
	# go back to event for time of day
	set stime [e startTime]
	set hour [clock format $stime -format {%H} ]
	set minute [clock format $stime -format {%M} ]
	set am "am"
	if { $hour > 12 } {
		set hour [expr $hour - 12]
		set am "pm"
	}
	if { $hour == "00" && $minute == "00" } {
		puts "[$e name]"
	} else {
		puts "$hour:$minute$am [$e name]"
	}
}

proc isException {e day } {
	global secondsperday
	set nd [$e nDateExceptions]
	if { $nd <= 0 } {
		return {}
	}
	for { set di 0} {$di < $nd} {incr di} {
		set x [$e exceptions $di]
		if { ($x < $day)&& ($day < [expr $x + $secondsperday]) } {
			return $x
		}
	}
	return {}
}


palm open datebook.dat myfile

# dumpPalmFileHeader myfile

set now [clock seconds]
set today [expr $now - $now % $secondsperday]

set basemonth [clock format $today -format {%m}]
set nextmonth [expr $basemonth + 1]
if { $nextmonth > 12 } {
	set nextmonth 1
}

for { set i 0 } { $i < [myfile numEntries] } { incr i } {
	myfile next e

	# puts "====== type [e type]"

	# puts "\nItem $i"
	set eventFlag [e eventFlag]
	set stime [e startTime]

	if { $stime > [expr $today + $secondsperday * 366 ] } {
		# skip things that are way off
		continue;
	}

	if { $eventFlag == 0 } {
		if { $stime > $today } {
			printMe e $stime
		}
		continue
	}

	# deal with repeats
	set end [e endDate]
	if { $end != $forever && $end < $today } {
		continue;
	}

	# brand is the repeat type
	set brand [e brand]
	set theName [e name]
	if { $brand == 5 } {
		# once a year
		puts -nonewline "[expr [e monthi] + 1] [e day]	"
		if { [regexp -nocase {birthday} "$theName"] } {
			puts -nonewline {+b }
		}
		if { [regexp -nocase {anniversay} $theName] } {
			puts -nonewline {+b }
		}
		puts "$theName"
	} elseif { $brand == 4 } {
		# once a month
		puts "$basemonth [e day] [e name]"
		puts "$nextmonth [e day] [e name]"
	} elseif { $brand == 2 } {
		# once a week
		set xday $today
		set daymask [e daymask]
		for { set xx 0 } \
		    { $xx < 366 } \
		    { incr xx ; incr xday $secondsperday } {
			if { $end != $forever && $end < $xday } {
				break
			}

			if { [isException e $xday] != {} } {
				continue
			}

			set dow [clock format $xday -format {%w}]
			if { $stime < $xday } {
				if { [expr ( 1 << $dow ) & $daymask] != 0 } {
					printMe e $xday
				}
			}
		}
	} elseif { $brand == 1 } {
		# daily
		set xday $today
		set daymask [e daymask]
		for { set xx 0 } \
		    { $xx < 366 } \
		    { incr xx ; incr xday $secondsperday } {
			if { $end != $forever && $end < $xday } {
				break
			}

			if { $stime < $xday } {
				printMe e $xday
			}

		}
	} else {
		# hope for the best
		printMe e $stime
		printRepeat e
	}

}

exit 0
