#!/usr/bin/wish

# test.tcl - test palm library for Tcl

# Get command line stuff
set files "$argv"
#puts "args = $files"

if { [ info exists VERBOSE ] == 0 } {
        set VERBOSE 0
        catch { set VERBOSE $env(VERBOSE) } 
}
if { [ info exists QUIET ] == 0 } {
        set QUIET 0
        catch { set QUIET $env(QUIET) } 
}
if { $QUIET > 0 } {
        set VERBOSE 0
}

# If we are in tclsh, this will fail and we will load the package.
# When linked as palmtcl, it is true so we don't load as a package.

if { [info commands palm ] == "" } {
	set x [ catch  {
		package require palm
	} msg ] 
	if { "$x" == 0 } {
		puts "loaded package: palm $msg"
	} else {
		puts "Could not load palm package"
		exit 1
	}
}

proc dumpPalmFileHeader { f } {

puts "dump palm file header"
	puts "#file: [$f filename]"
	puts "#signature: [format {0x%x} [$f signature]] = [$f signatureName]"
	puts "#custheader: [$f custheader]"
	puts "#nextFreeCategory: [$f nextFreeCategory]"
	puts "#categoryCount: [$f categoryCount]"

	for { set i 1 } { $i <= [$f categoryCount] } { incr i } {
		puts "category $i: [$f category $i]"
	}

	puts "#resourceID: [$f resourceID]"
	puts "#fieldsPerRow: [$f fieldsPerRow]"
	puts "#recordIDPos: [$f recordIDPos]"
	puts "#recordStatusPos: [$f recordStatusPos]"
	puts "#placementPos: [$f placementPos]"

	puts "#fieldCount: [$f fieldCount]"
	for { set i 1 } { $i <= [$f fieldCount] } { incr i } {
		puts "#field\[[expr $i - 1]\]: [$f fieldtype $i]"
	}
	puts "#numEntries: [$f numEntries]"
}


proc cat2text { file catIndex } {
	if { $catIndex <= 0 } {
		return "unknown"
	}
	return "[$file category $catIndex] ([$file category $catIndex id])"
}

proc printIf { tag text } {
	if { "$text" != "" } {
		puts -nonewline "$tag$text"
	}
}

proc printIfln { tag text } {
	if { "$text" != "" } {
		if { "$tag" != "" } { puts -nonewline "$tag: " }
		puts "$text"
	}
}


proc bool2text { value tag } {
	if { $value == "0" } { return "" }
	if { $value == "1" } { return " $tag" }
	return " $tag(0x[format %x $value])"
}

proc time2str { timeval } {
	if { $timeval == 0x749d3450 } {
		return "forever";
	}
	return "[clock format $timeval -format {%Y-%02m-%02d %H:%M:%S} -gmt 1]=$timeval"
}

proc todo_print {e} {
	puts -nonewline "[$e description]"
	puts -nonewline ", priority: [$e priority]"
	puts -nonewline ", completed: [$e completed]"
	puts "due: [clock format [$e due] -format {%Y-%m-%d %H:%M:%S} ]"
	if { [$e note] != "" } {
		puts "note {[$e note]}"
	}
}

proc memo_print {e} {
	puts "[$e memo]"
}

proc address_print {e} {

	if { [$e lastname] != "" } {
		puts -nonewline "[$e lastname]"
		if { [$e firstname] != "" } {
			puts -nonewline ", "
		}
	}
	printIf "" "[$e firstname]"
	puts ""

	printIfln company "[$e company]"
	printIfln title "[$e title]"
	printIfln "" "[$e address]"
	printIf "" "[$e city]"
	printIf "," "[$e state]"
	printIf "  " "[$e zip]"
	printIf ", country" "[$e country]"
	puts ""

	for { set i 0 } { $i < 5 } { incr i } {
		set pi [expr $i + 1]
		puts -nonewline "phone $i "

		if { [$e displayIndex] == $pi } {
			puts -nonewline "*"
		} else {
			puts -nonewline " "
		}
		puts "[$e phone $i label] [$e phone $i]"
	}
	if { [$e note] != "" } {
		puts "note {[$e note]}"
	}
}

proc printRepeat { e } {

	set eventFlag [$e eventFlag]
	if { $eventFlag == 0 } return

	puts -nonewline "repeat {"
	if { $eventFlag == 0xffff } {
		puts -nonewline "class '[$e classname]'"
	} else {
		puts -nonewline "[format {0x%x} $eventFlag]"
	}
	set brand [$e brand]

	puts -nonewline " brand 0x[format %x $brand] interval [$e interval]"
	if { [$e endDate] != 0 } {
		puts -nonewline " enddate [time2str [$e endDate]]"
	}
	puts -nonewline " dow [$e dayOfWeek]"

	# brand data
	if { $brand <= 3 } {
		puts -nonewline " day [$e day]"
	} 
	if { $brand == 2 } {
		puts -nonewline " daymask [$e daymask]"
	}
	if { $brand == 3 } {
		puts -nonewline " weeki [$e weeki]"
	}
	if { $brand == 4 || $brand == 5 } {
		puts -nonewline " day# [$e day]"
	}
	if { $brand == 5 } {
		puts -nonewline " monthi [$e monthi]"
	}
	puts "}";
	if { [$e nDateExceptions] > 0 } {
		puts -nonewline "exceptions {"
		for { set i 0} {$i < [$e nDateExceptions]} {incr i} {
			puts -nonewline " [$e exceptions $i]"
		}
		puts "}"
	}
}

proc date_print {e} {
	puts -nonewline "{[$e name]}"

	puts -nonewline " start [clock format [$e startTime] -format {%Y-%m-%d %H:%M:%S} -gmt 1]"
	puts -nonewline " end [clock format [$e endTime] -format {%Y-%m-%d %H:%M:%S} -gmt 1]"

	puts -nonewline [bool2text [$e untimed] "untimed"]
	puts -nonewline [bool2text [$e alarm] "alarm"]

	if { [$e alarm] != 0 && [$e alarmWarn] != 0 } {
		puts -nonewline " warn [$e alarmWarn]"
		if { [$e alarmWarnType] == 0 } {
			puts -nonewline "m"
		} elseif { [$e alarmWarnType] == 1 } {
			puts -nonewline "h"
		} elseif { [$e alarmWarnType] == 2 } {
			puts -nonewline "d"
                } else {
			puts -nonewline "{illegal warntype [$e alarmWarnType]}"
                }
        }
	puts ""

	printRepeat $e

	if { [$e note] != "" } {
		puts "note {[$e note]}"
	}
}

puts "palm library info:"
puts "build date: [palm build_date]"
puts "revision date: [palm revision_date]"

#palm open tests/todo.dat myfile
#palm open tests/memopad.dat myfile
#palm open tests/address.dat myfile
#palm open tests/datebook.dat myfile
palm open HOLIDAYS.DBA myfile

dumpPalmFileHeader myfile

for { set i 0 } { $i < [myfile numEntries] } { incr i } {
	myfile next e
	# puts "====== type [e type]"

	puts "\nItem $i"
	puts -nonewline "id: [format {%08x} [e id]]"
	puts -nonewline ", status: [format {0x%x} [e status]]"
	if { [e position] != "" } {
		puts -nonewline ", position: [format %x [e position]]"
	}
	if { [e isPrivate] != 0 } {
		puts -nonewline ", Private"
	}
	puts ", category: [cat2text myfile [e categoryIndex]]"

        set err [catch {
		set printProc "[e type]_print"
		# puts "proc is $printProc"
		if { [info commands $printProc ] != "" } {
			$printProc e
		} else {
			puts "no printer"
		}
	} errMsg ]
	if { $err != 0 } {
		puts "error $err = $errMsg"
	}
}
