/* palmtcl.cpp - Tcl integration with palm library
 *
 * This module may be compiled conditionally to make a loadable
 * Tcl module, or to make a stand alone interpreter application.
 * To build as an application make sure that LINK_AS_TCL_APP is
 * defined to a non-zero value
 */

/*
 * Copyright (c) 2002 Tony Aiuto
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * Author: Tony Aiuto (tony@aiu.to)
 *
 * Edit History:
 *
 * $Log: palmtcl.cpp $
 * Revision tools_TonyAiuto/2 2002/09/07 05:48:48 TonyAiuto
 * 	add copyright stuff
 * 
 * Wed Sep  4 23:59:18 EDT 2002
 * tony - created
 */

#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 2002, Tony Aiuto";
static char rcsid[] = "@(#) $Id: palmtcl.cpp tools_TonyAiuto/2 2002/09/07 05:48:48 TonyAiuto $";
#endif


#define	DEBUG	0

#define	EDIT_DATE "2002.09.04"		/* change on each edit */
#if !defined(BUILD_DATE)
#define	BUILD_DATE EDIT_DATE
#endif

#define	PALM_PKG_VERSION	"1.0"


#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <assert.h>

#if REMOVED
#if defined(WIN32)
#include <winsock2.h>
#else
#include <unistd.h>
#endif

#include <wchar.h>
#endif /* REMOVED */

#include <palmlib.h>
#include <palmtcl.h>


#if defined(WIN32)
#if defined(BUILD_Pun)
#undef TCL_STORAGE_CLASS
#define TCL_STORAGE_CLASS DLLEXPORT
#else
#undef EXTERN
#define EXTERN /* */
#endif /* BUILD_Pun */
#endif /* WIN32 */



/* Globals */
static PalmCommand *thePalmCommand;

static int extractMember(PalmRecordCommand *, Tcl_Interp *interp, int argc, char *argv[]);

static void usage()
{
	printf("usage: palmtcl {options} Tcl_Script_File args ...\n");

	printf("options: [-V] [-v] [-e expression] ... \n");

	printf("palm commands\n");
	printf("\tpalm build_date\t\t\treturn build date of module\n");
	printf("\tpalm revision_date\t\treturn last edit date of module\n");
	printf("\tpalm verbose\t\t\treturn verbosity level\n");
	printf("\tpalm verbose N\t\t\tset verbosity level\n");

	printf("\tpalm open datafile file_id\topen datafile as file_id\n");

	printf("\n");
	printf("    file commands\n");
	printf("\tfile_id next entry_id\t\tget next entry as entry_id\n");


	printf("\tfile_id categoryCount\t\t# of categories\n");
	printf("\tfile_id custheader\t\tcustom header\n");
	printf("\tfile_id fieldCount\t\t# of field types in schema\n");
	printf("\tfile_id fieldsPerRow\t\t# of fields per record\n");
	printf("\tfile_id filename\t\tname embedded within the file\n");
	printf("\tfile_id nextFreeCategory\t\t\n");
	printf("\tfile_id numEntries\t\t# of records in file\n");

	printf("\tfile_id placementPos\t\tfield index where placement is stored\n");
	printf("\tfile_id recorddIDPos\t\tfield index where record ID is stored\n");
	printf("\tfile_id recorddStatusPos\tfield index where rec. status is stored\n");

	printf("\tfile_id resourceID\t\tresource ID\n");
	printf("\tfile_id signature\t\tfile signature\n");
	printf("\tfile_id category N\t\tname of Nth category\n");
	printf("\tfile_id fieldType N\t\ttype of Nth field\n");

	printf("\n");
	printf("    entry commands\n");
	printf("\tentry_id type\t\trecord type {address, date, memo, todo}\n");
	printf("\tentry_id FIELD\t\tretrieve FIELD, where FIELD is one of:\n");

	printf("\t\tall:\t\tid, position, status, categoryIndex\n");
	printf("\t\taddress:\tfirstname, lastname, company, title\n");
	printf("\t\t\t\taddress, city, state, zip, country\n");
	printf("\t\t\t\tnote, displayIndex\n");
	printf("\t\tdate:\t\talarm alarmWarn alarmWarnType duration\n");
	printf("\t\t\t\tendTime name note startTime untimed\n");
	printf("\t\tmemo:\t\tmemo\n");
	printf("\t\ttodo:\t\tcompleted, description, due, priority, note\n");

	printf("\taddress_id phone N {index label}\t\tretrieve phone #N\n");
	printf("\t\t\t\tOptionally get index or type of number\n");
	printf("\taddress_id custom N\tretrieve custom field N\n");


	printf("\nIn the above, *_id always refers to a TCL identifier, not an actual key\n");
	exit(0);
}


/************************  Utility functions ****************************/




/****************************************************************************/

/* PalmCommand
 * This method deals with generic Palm things
 */

PalmCommand::PalmCommand(Tcl_Interp *interp) : TCLCommand(interp, "palm") { }


/* command
 *
 * Arguments:
 * clientData : Not used
 * interp     : Pointer to the Tcl interpreter
 * argc       : Number of arguments passed to "palm" in Tcl
 * argv       : Values of the arguments passed to "palm" in Tcl
 *		--> argv[1] - command
 *		samples:
 *			build_date
 *			revision_date
 *			verbose [N]
 *			open FILENAME FILE_ID
 *
 * Return value:
 * Error      --> TCL_ERROR
 * Otherwise  --> TCL_OK
 */

int PalmCommand::command(Tcl_Interp *interp, int argc, char *argv[])
{

	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;
	     GETTER_S("build_date", BUILD_DATE)
	else GETTER_S("revision_date", EDIT_DATE)
	else if(strcmp(argv[1], "verbose") == 0) {
		if(argc <= 2) {
			setResult(interp, getVerbose());
		} else {
			setVerbose( atoi(argv[2]) );
		}
	} else if(strcmp(argv[1], "open") == 0) {
		if(TCLutil::checkArgs(interp, argc, argv, 3)) return TCL_ERROR;

		// palm open filename reader_id
		const char *fname = argv[2];
		const char *id = argv[3];
		FILE *inp = fopen(fname, "r");
		if(inp == NULL) {
			char *tmp = Tcl_Alloc(strlen(fname) + 30);
			sprintf(tmp, "Cannot open '%s'", fname);
			Tcl_SetResult(interp, tmp, TCL_DYNAMIC);
			return TCL_ERROR;
		}

		// Note: PalmRead takes ownership of inp
		PalmRead *reader = new PalmRead(inp);
		assert(reader != NULL);
		reader->loadheader();

#if DEBUG
		reader->printheader(out);
#endif

		// we do not care about the object - TclCommand frees it
		new PalmReaderCommand(interp, id, reader);

	} else {
		return TCL_ERROR;
	}

	return TCL_OK;
}



/****************************************************************************/


PalmReaderCommand::PalmReaderCommand(Tcl_Interp *interp, const char *_name, PalmRead *_reader)
    :
	TCLCommand(interp, _name),
	reader(_reader)
{
	// printf("PalmReaderCommand::ctor(%s)\n", name);
}

PalmReaderCommand::~PalmReaderCommand()
{
	// printf("PalmReaderCommand::dtor(%s)\n", name);
	if(reader != NULL) delete reader;
}



/* PalmReaderCommand command proc
 *
 * Arguments:
 * interp     : Pointer to the Tcl interpreter
 * argc       : Number of arguments passed 
 * argv       : Values of the arguments passed
 *		--> argv[1] - method name
 *
 * Return value:
 * Error      --> TCL_ERROR, interpreter result contains error msg
 * Otherwise  --> TCL_OK, interpreter result contains some result
 */

int  PalmReaderCommand::command(Tcl_Interp *interp, int argc, char *argv[])
{
	// printf("PalmReader command\n");

	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;

	     GETTER_I("categoryCount", reader->getCategoryCount())
	else GETTER_S("custheader", reader->getCustheader())
	else GETTER_I("fieldCount", reader->getFieldCount())
	else GETTER_I("fieldsPerRow", reader->getFieldsPerRow())
	else GETTER_S("filename", reader->getFilename())
	else GETTER_I("nextFreeCategory", reader->getNextFreeCategory())
	else GETTER_I("numEntries", reader->getNumEntries())
	else GETTER_I("placementPos", reader->getPlacementPos())
	else GETTER_I("recordIDPos", reader->getRecordIDPos())
	else GETTER_I("recordStatusPos", reader->getRecordStatusPos())
	else GETTER_I("resourceID", reader->getResourceID())
	else GETTER_I("signature", reader->getSignature())
	else if(strcmp(argv[1], "signatureName") == 0) {
		long fileSig = reader->getSignature();
		if(fileSig == PALM_DBA_SIG) {
			Tcl_SetResult(interp, "datebook", TCL_STATIC);
		} else if(fileSig == PALM_ABA_SIG) {
			Tcl_SetResult(interp, "address", TCL_STATIC);
		} else if(fileSig == PALM_MEMO_SIG) {
			Tcl_SetResult(interp, "memo", TCL_STATIC);
		} else if(fileSig == PALM_TODO_SIG) {
			Tcl_SetResult(interp, "todo", TCL_STATIC);
		} else {
			Tcl_SetResult(interp, "unknown", TCL_STATIC);
		}
	} else if(strcmp(argv[1], "category") == 0) {
		char tmp[50];
		if(TCLutil::checkArgs(interp, argc, argv, 2)) return TCL_ERROR;

		int ci = atoi(argv[2]);
		if(ci < 1 || ci > reader->getCategoryCount()) {
			sprintf(tmp, "category index (%d) out of range (1..%d)",
					ci, reader->getCategoryCount());
			Tcl_SetResult(interp, tmp, TCL_VOLATILE);
			return TCL_ERROR;
		}
		PalmCategory *c = reader->getCategory(ci-1);
		char *ret = c->longname;
		if(argc > 3) {
			if(strcmp(argv[3], "short") == 0) {
				ret = c->shortname;
			} else if(strcmp(argv[3], "id") == 0) {
				sprintf(tmp,"%d", c->id);
				ret = tmp;
			} else if(strcmp(argv[3], "dirty") == 0) {
				sprintf(tmp,"%d", c->dirty);
				ret = tmp;
			}
		}
		Tcl_SetResult(interp, ret, TCL_VOLATILE);
	} else if(strcmp(argv[1], "fieldtype") == 0) {
		if(TCLutil::checkArgs(interp, argc, argv, 2)) return TCL_ERROR;
		char tmp[50];

		int fi = atoi(argv[2]);
		if(fi < 1 || fi > reader->getFieldCount()) {
			sprintf(tmp, "field index (%d) out of range (1..%d)",
					fi, reader->getFieldCount());
			Tcl_SetResult(interp, tmp, TCL_VOLATILE);
			return TCL_ERROR;
		}
		setResult(interp, reader->fieldType(fi-1));
	} else if(strcmp(argv[1], "next") == 0) {
		if(TCLutil::checkArgs(interp, argc, argv, 2)) return TCL_ERROR;

		PalmEntry *e = reader->getEntry();
		// we do not care about the object - TclCommand frees it
		new PalmRecordCommand(interp, argv[2], e, reader);

	} else {
		// Pass back unknown things
		Tcl_SetResult(interp, argv[1], TCL_VOLATILE);
	}

	return TCL_OK;
}


/****************************************************************************/


PalmRecordCommand::PalmRecordCommand(Tcl_Interp *interp, const char *_name, PalmEntry *_entry, PalmRead *_reader)
    :
	TCLCommand(interp, _name),
	entry(_entry),
	myfile(_reader)
{
	// printf("PalmRecordCommand::ctor(%s)\n", name);
}

PalmRecordCommand::~PalmRecordCommand()
{
	// printf("PalmRecordCommand::dtor(%s)\n", name);
	if(entry != NULL) delete entry;
}


/* PalmRecordCommand command proc
 *
 * Arguments:
 * interp     : Pointer to the Tcl interpreter
 * argc       : Number of arguments passed 
 * argv       : Values of the arguments passed
 *		--> argv[1] - method name
 *
 * Return value:
 * Error      --> TCL_ERROR, interpreter result contains error msg
 * Otherwise  --> TCL_OK, interpreter result contains some result
 */

int  PalmRecordCommand::command(Tcl_Interp *interp, int argc, char *argv[])
{
	char	tmp[50];
	// printf("PalmRecord command\n");

	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;

	     GETTER_I("id", entry->recordID)
	else GETTER_I("categoryIndex", entry->categoryIndex)
	else GETTER_I("isPrivate", entry->isPrivate)
	else if(strcmp(argv[1], "position") == 0) {
		if(entry->position != INT_MAX) {
			setResult(interp, entry->position);
		} else {
			Tcl_SetResult(interp, "", TCL_STATIC);
		}
	} else GETTER_I("status", entry->recStatus)
	else if(strcmp(argv[1], "type") == 0) {
		// NOTE : cast is save because of VOLATILE
		Tcl_SetResult(interp, (char *) entry->type(), TCL_VOLATILE);
	} else if(extractMember(interp, argc, argv) == TCL_OK) {
		// done
		;
	} else {
		// Pass back unknown things
		Tcl_SetResult(interp, argv[1], TCL_VOLATILE);
	}
	return TCL_OK;
}



/****************************************************************************/

int PalmRecordCommand::extractDate(PalmEntry *_e, Tcl_Interp *interp, int argc,char *argv[])
{
	DateBookEntry *e = (DateBookEntry *) _e;
	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;

	     GETTER_I("alarm", e->alarm)
	else GETTER_I("alarmWarn", e->alarmWarn)
	else GETTER_I("alarmWarnType", e->alarmWarnType)
	else GETTER_I("duration", e->duration)
	else GETTER_I("endTime", e->endTime)
	else GETTER_S("name", e->name)
	else GETTER_S("note", e->note)
	else GETTER_I("startTime", e->startTime)
	else GETTER_I("untimed", e->untimed)

	// Repeat block things
	else GETTER_S("classname", e->repeat.classname)
	else GETTER_I("nDateExceptions", e->repeat.nDateExceptions)
	else GETTER_I("eventFlag", e->repeat.eventFlag)
	else GETTER_I("brand", e->repeat.brand)
	else GETTER_I("interval", e->repeat.interval)
	else GETTER_I("endDate", e->repeat.endDate)
	else GETTER_I("dayOfWeek", e->repeat.dayOfWeek)
	else GETTER_I("day", e->repeat.day)
	else GETTER_I("daymask",e->repeat.daymask)
	else GETTER_I("weeki", e->repeat.weeki)
	else GETTER_I("monthi", e->repeat.monthi)

	else if(strcmp(argv[1], "exceptions") == 0) {
		if(TCLutil::checkArgs(interp, argc, argv, 2)) return TCL_ERROR;
		setResult(interp, e->repeat.dateException(atoi(argv[2])));
	} else {
		return ~TCL_OK;
	}


	return TCL_OK;
}

int PalmRecordCommand::extractAddress(PalmEntry *_e, Tcl_Interp *interp,int argc,char *argv[])
{
	AddressBookEntry *e = (AddressBookEntry *) _e;
	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;

	     GETTER_S("company", e->company)
	else GETTER_S("firstname", e->firstname)
	else GETTER_S("lastname", e->lastname)
	else GETTER_S("title", e->title)

	else GETTER_S("address", e->address)
	else GETTER_S("city", e->city)
	else GETTER_S("state", e->state)
	else GETTER_S("zip", e->zip)
	else GETTER_S("country", e->country)

	else GETTER_I("displayIndex", e->displayIndex)
	else GETTER_S("note", e->note)
	else if(strcmp(argv[1], "phone") == 0) {
		char tmp[20];

		if(TCLutil::checkArgs(interp, argc, argv, 2)) return TCL_ERROR;
		int pi = atoi(argv[2]);
		if(pi >= 5) pi = 0;
		int ptype = 0;
		if(argc > 3) {
			if(strcmp(argv[3], "index") == 0) ptype = 1;
			if(strcmp(argv[3], "label") == 0) ptype = 2;
		}
		if(ptype == 0) {
			Tcl_SetResult(interp, e->phoneNumbers[pi]->text,
						TCL_VOLATILE);
		} else if(ptype == 1) {
			setResult(interp, e->phoneNumbers[pi]->labelIndex);
		} else {
			int li = e->phoneNumbers[pi]->labelIndex;
			char *ptype;
			if(0 <= li && li < PalmUtil::nPhoneTypes()) {
				Tcl_SetResult(interp, 
					PalmUtil::phoneTypes[li], TCL_VOLATILE);
			} else {
				sprintf(tmp, "(*** %d ***)", li);
				ptype = tmp; 
			}
		}
	} else if(strcmp(argv[1], "custom") == 0) {
		if(TCLutil::checkArgs(interp, argc, argv, 2)) return TCL_ERROR;
		int pi = atoi(argv[2]);
		if(pi >= 4) pi = 0;
		Tcl_SetResult(interp, e->custom[pi], TCL_VOLATILE);
	} else {
		return ~TCL_OK;
	}

	return TCL_OK;
}

int PalmRecordCommand::extractMemo(PalmEntry *_e, Tcl_Interp *interp,int argc,char *argv[])
{
	MemoEntry *e = (MemoEntry *) _e;
	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;

	GETTER_S("memo", e->memo)
	else {
		return ~TCL_OK;
	}

	return TCL_OK;
}

int PalmRecordCommand::extractTodo(PalmEntry *_e, Tcl_Interp *interp,int argc,char *argv[])
{
	TodoEntry *e = (TodoEntry *) _e;
	// Check for the correct # of arguments.
	if(TCLutil::checkArgs(interp, argc, argv, 1)) return TCL_ERROR;

	     GETTER_I("completed", e->completed)
	else GETTER_S("description", e->description)
	else GETTER_I("due", e->due)
	else GETTER_I("priority", e->priority)
	else GETTER_S("note", e->note)
	else {
		return ~TCL_OK;
	}

	return TCL_OK;
}


/* This is a terribly ugly thing.  What we want is a polymorphic
 * dispatch.  We could write a "field name to string value" routine
 * as part of the PalmEntry class and have each subclass implement
 * it, but that would be pushing too much down into the palm library
 *
 * Better, would be to have the getEntry foundary be able to return
 * a caller specified subclass of WhateverEntry which implements that,
 * but that ain't happening in C++ in my lifetime.
 *
 */

int PalmRecordCommand::extractMember(Tcl_Interp *interp,
			int argc, char *argv[])
{
	int ret = TCL_OK;
	if(myfile->getSignature() == PALM_DBA_SIG) {
		ret = extractDate(entry, interp, argc, argv);
	} else if(myfile->getSignature() == PALM_ABA_SIG) {
		ret = extractAddress(entry, interp, argc, argv);
	} else if(myfile->getSignature() == PALM_MEMO_SIG) {
		ret = extractMemo(entry, interp, argc, argv);
	} else if(myfile->getSignature() == PALM_TODO_SIG) {
		ret = extractTodo(entry, interp, argc, argv);
	} else {
		fprintf(stderr, "Internal problem.  We have a type which we can't handle.\n");
	}
	return ret;
}



/****************************************************************************/


/* Palm_Init
 * Add Palm related commands to TCL
 *
 * Arguments:
 * interp   : Pointer to the Tcl interpreter to be initialized.
 *
 * Return value:
 * TCL_OK      : If everything goes fine
 * TCL_ERROR   : If any of the initializations fail
 *
 * Side effects:
 *	The Palm package is created.
 *	new commands added to the Tcl interpreter:  palm
 */

/* EXTERN linkage comes from tcl.h, does the right thing for windows */
extern "C" EXTERN int Palm_Init(Tcl_Interp *interp)
{
#if LINK_AS_TCL_APP == 0

	/* Build as a Tcl extension */

	if (Tcl_InitStubs(interp, "8.3", 0) == NULL) {
		return TCL_ERROR;
	}
	if (Tcl_PkgRequire(interp, "Tcl", TCL_VERSION, 0) == NULL) {
		if (TCL_VERSION[0] == '7') {
			if (Tcl_PkgRequire(interp, "Tcl", "8.0", 0) == NULL) {
				return TCL_ERROR;
			}
		}
	}
	if (Tcl_PkgProvide(interp, "palm", PALM_PKG_VERSION) != TCL_OK) {
		return TCL_ERROR;
	}
#else

	// here for standalone app

#if defined(unix)
	// This seems to break under windoze
	if (Tcl_Init(interp) == TCL_ERROR) {
 		printf("Tcl_Init returns error\n");
		return TCL_ERROR;		
  	}
#endif

#endif

	thePalmCommand = new PalmCommand(interp);

	return TCL_OK;
}



#if LINK_AS_TCL_APP

#if defined(WIN32) || defined(sun)
extern "C" {
extern int getopt(...);
extern int optind;
extern char *optarg;
}
#else
#include <getopt.h>
#endif

int main(int argc, char *argv[])
{
	int	c;
	char	*emailed = NULL;
	Tcl_Interp	*interp;
	char	*evals[30];		// expressions to pre-evaluate
	int	nevals = 0;
	int	verbose;

	// printf("argc = %d, argv[0] = %s\n", argc, argv[0]);

	while((c = getopt(argc, argv, "e:P:vV")) != EOF) {
		switch(c) {
		case 'e':
			if(nevals < sizeof(evals)/sizeof(evals[0]) - 1) {
				evals[nevals++] = strdup(optarg);
			} else {
				fprintf(stderr, "Too many -e expressions\n");
				exit(1);
			}
			break;

		case 'v':
			verbose++;
			break;

		case 'V':
			printf("palmtcl %s\n" , PALM_PKG_VERSION);
			printf("Revision date: %s\n", EDIT_DATE);
			printf("build date: %s\n", BUILD_DATE);
			exit(0);

		default:
			usage();
		}
	}
	if(argc - optind < 1) {
		usage();
	}

	interp = Tcl_CreateInterp();
	Palm_Init(interp);
	if(verbose) thePalmCommand->setVerbose(verbose);

	/* evaluated -e options */
	for(int ni = 0; ni < nevals; ni++) {
		if(verbose) printf("Eval(%s)\n", evals[ni]);
		Tcl_Eval(interp, evals[ni]);
	}

	/* provide trailing args */
	// printf("optind=%d, argc=%d\n", optind, argc);
	Tcl_SetVar(interp, "argv",
		Tcl_Merge(argc - optind - 1, &argv[optind+1]),
		TCL_GLOBAL_ONLY);

	int code = Tcl_EvalFile(interp, argv[optind]);
	if(*interp->result != 0) {
		printf("%s\n", interp->result);
	}

	if(code != TCL_OK) {
		return 1;
	}
	return 0;
}


#endif	/* LINK_AS_TCL_APP */


/* End of palmtcl.cpp */
