/* tclutil.cpp - tcl utility functions
 *
 * This is a work-in-progress utility set for making it easier to
 * integrate software with TCL.
 */

/*
 * Copyright (c) 2002 Tony Aiuto
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor their employers are responsible for
 *    any consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors must appear
 *    in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This copyright notice may not be removed or altered.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL TONY AIUTO, HIS CURRENT OR PAST EMPLOYERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


/*
 * Author: Tony Aiuto (tony@aiu.to)
 * Edit History:
 *
 * $Log: tclutil.cpp $
 * Revision tools_TonyAiuto/2 2002/09/07 05:48:48 TonyAiuto
 * 	add copyright stuff
 * 
 * Wed Sep  4 23:52:32 EDT 2002
 * tony - created
 */

#ifndef lint
static char _copyright[] = "@(#) Copyright (c) 2002, Tony Aiuto";
static char rcsid[] = "@(#) $Id: tclutil.cpp tools_TonyAiuto/2 2002/09/07 05:48:48 TonyAiuto $";
#endif


#include <tcl.h>
#include <tclutil.h>


/** Check for requisite number of args
 */
int TCLutil::checkArgs(Tcl_Interp *interp, int argc, char *argv[], int expect)
{
	// printf("checkargs, argc=%d, expect=%d\n", argc, expect);
	if(argc < expect+1) {
		printf("Wrong # of args:");
		for(int i = 0; i < argc; i++) {
			printf(" %s", argv[i]);
		}
		printf("\n");
		Tcl_SetResult(interp, "Wrong # of args.", TCL_STATIC);
		return TCL_ERROR;
	}
	return 0;
}




/** TCLCommand *********************************************************/

TCLCommand::TCLCommand(Tcl_Interp *interp, const char *commandname)
{
	name = new char [strlen(commandname) + 1];
	strcpy(name, commandname);

#if defined(DEBUG) && DEBUG > 1
	printf("TCLCommand::TCLCommand(%s)\n", name);
#endif

	Tcl_CreateCommand(interp, name, &commandDispatch, (ClientData) this,
						&cleanupDispatch);
}


TCLCommand::~TCLCommand()
{
#if defined(DEBUG) && DEBUG > 1
	printf("TCLCommand::~TCLCommand()\n");
#endif
	delete [] name;
}

int TCLCommand::commandDispatch(ClientData clientData, Tcl_Interp *interp,
				int argc, char *argv[])
{
	TCLCommand *self = (TCLCommand *) clientData;
#if DEBUG > 1
	printf("Dispatching to command proc\n");
#endif
	self->command(interp, argc, argv);
}

void TCLCommand::cleanupDispatch(ClientData clientData)
{
	TCLCommand *self = (TCLCommand *) clientData;
	delete self;
}


void TCLCommand::setResult(Tcl_Interp *interp, long i)
{
	char tmp[80];
	sprintf(tmp, "%ld", i);
	Tcl_SetResult(interp, tmp, TCL_VOLATILE);
}

